# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import re

PERCENT_MATCHER = re.compile("%([^%])?")


def escape_percentage(value: str) -> str:
    return PERCENT_MATCHER.sub(r"%%\1", value)



# SPDX-FileCopyrightText: 2023 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import abc
import argparse
import logging
from collections.abc import Callable
from pathlib import Path
from typing import Any

fmt = "{levelname}:{name}: {message}"
logging.basicConfig(format=fmt, style="{")
LOG = logging.getLogger("fclogr")


class InvalidArgumentError(Exception):
    """
    A problem parsing or validating a command line argument
    """


class Command(abc.ABC):
    _cleanup: list[Callable[[], Any]]

    @classmethod
    @abc.abstractmethod
    def make_parser(
        cls, parser_func: Callable = argparse.ArgumentParser, standalone=False, **kwargs
    ) -> argparse.ArgumentParser: ...

    @abc.abstractmethod
    def run(self) -> str | int: ...

    @property
    def cleanup(self) -> list[Callable[[], Any]]:
        if getattr(self, "_cleanup", None) is None:
            self._cleanup = []
        return self._cleanup

    def _v_specpath(self, path: Path | None = None) -> Path:
        if not path:
            pwd = Path.cwd()
            path = Path(pwd.name + ".spec")
        if not path.is_file() or path.suffix != ".spec":
            raise InvalidArgumentError(f"{path} must exist an end with '.spec'")
        return path



# SPDX-FileCopyrightText: 2023 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import argparse
import logging
import sys
from collections.abc import Sequence
from subprocess import CalledProcessError

from fclogr.cli.sync import Syncer

from .base import LOG, Command, InvalidArgumentError
from .bump import Bumper
from .dev_entries import DevEntries, DevSRPM


def get_command(argv: Sequence[str] | None = None, **kwargs) -> Command:
    parser = argparse.ArgumentParser(
        description="fclogr is a tool for managing RPM changelogs and updates", **kwargs
    )
    log_group = parser.add_mutually_exclusive_group()
    log_group.add_argument(
        "--debug", action="store_const", dest="log_level", const="DEBUG"
    )
    log_group.add_argument(
        "--ll",
        "--log-level",
        dest="log_level",
        type=lambda t: t.upper(),
        choices=["DEBUG", "INFO"],
    )
    subparsers = parser.add_subparsers(
        title="Subcommands", dest="action", required=True
    )
    for name, cls in COMMANDS.items():
        cls.make_parser(subparsers.add_parser, standalone=False, name=name)

    args = vars(parser.parse_args(argv))
    action = args.pop("action")
    if log_level := args.pop("log_level"):
        level = getattr(logging, log_level)
        LOG.setLevel(level)

    return COMMANDS[action](**args)


def main(argv: Sequence[str] | None = None) -> int | str:
    try:
        command = get_command(argv)
    except InvalidArgumentError as err:
        # sys.exit() in the console_script handles this
        return str(err)
    try:
        r = command.run()
    except CalledProcessError as err:
        print(err, file=sys.stderr)
        r = err.returncode
    except Exception:
        LOG.exception("Unexpected Exception")
        r = 1
    for func in command.cleanup:
        func()
    return r


COMMANDS: dict[str, type[Command]] = {
    "bump": Bumper,
    "dev-entries": DevEntries,
    "dev-srpm": DevSRPM,
    "sync": Syncer,
}



# Copyright (c) 2023 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import shutil

import pytest

from fclogr.cli import main
from fclogr.cli.base import InvalidArgumentError
from fclogr.cli.sync import Syncer


def test_syncer_same(test_data, tmp_path):
    data = tmp_path / "data"
    shutil.copytree(test_data, data)
    path = data.joinpath("package.spec")
    with pytest.raises(
        InvalidArgumentError, match="Input and output are the same file"
    ):
        Syncer([path], [path])


def test_syncer(test_data, tmp_path):
    data = tmp_path / "data"
    shutil.copytree(test_data, data)
    assert main(["sync", str(data / "package.spec"), str(data / "package.2.spec")]) == 0
    # ruff: noqa: SIM117
    with open(data / "package.2.spec") as output:
        with open(data / "package.3.spec") as expected:
            for o, e in zip(output, expected):
                assert o == e



# SPDX-FileCopyrightText: 2023 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import fclogr.cli

if __name__ == "__main__":
    fclogr.cli.main()



# SPDX-FileCopyrightText: 2023 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import argparse
import datetime as dt
import functools
import logging
import shutil
import subprocess
import sys
import tarfile
import tempfile
from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING

import pygit2
import specfile as sfile
from specfile import options

from .._util import escape_percentage
from .base import Command, InvalidArgumentError

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath

LOG = logging.getLogger(__name__)


class GitLogEntry:
    def __init__(
        self,
        *,
        commit: pygit2.Commit,
        baseversion: str,
        pre: bool,
        index: str | int = 1,
    ):
        self.commit = commit
        self.baseversion = baseversion
        self.pre = pre
        self.index = index

    @property
    def message(self) -> list[str]:
        message = self.commit.message.splitlines()
        # --allow-empty-message
        if not message:
            message = ["bump"]
        return ["- " + escape_percentage(message[0])]

    @property
    def author(self) -> str:
        return f"{self.commit.author.name} <{self.commit.author.email}>"

    @property
    def date(self) -> dt.date:
        # Use UTC date
        timestamp = dt.datetime.fromtimestamp(
            self.commit.commit_time, tz=dt.timezone.utc
        )
        return timestamp.date()

    @property
    def version(self) -> str:
        return "{ref}{sep}{num}.{date:%Y%m%d}.{hash}".format(
            ref=self.baseversion,
            sep="~" if self.pre else "^",
            num=self.index,
            date=self.date,
            hash=self.commit.short_id,
        )

    @property
    def evr(self) -> str:
        return self.version + "-1"

    @property
    def entry(self) -> sfile.changelog.ChangelogEntry:
        return sfile.changelog.ChangelogEntry.assemble(
            self.date, self.author, self.message, self.evr
        )

    def add_to_clog(self, clog: sfile.changelog.Changelog) -> None:
        if self.entry not in clog:
            clog.append(self.entry)


class DevEntries(Command):
    baseversion: str
    pre: bool

    def __init__(
        self,
        *,
        specpath: Path | None,
        outdir: Path,
        stdout: bool = False,
        archive: bool,
        baseversion: str | None,
        pre: str | None,
        evr_only: bool = False,
    ) -> None:
        self.specpath = self._v_specpath(specpath)

        self.outdir: Path = outdir
        if not self.outdir.is_dir():
            raise InvalidArgumentError(f"--outdir '{self.outdir}' is not a directory")

        self.stdout: bool = stdout

        self.archive: bool = archive

        self.git_path: Path = Path.cwd()

        try:
            self.spec = sfile.Specfile(self.specpath)
        except sfile.exceptions.SpecfileException as err:
            raise InvalidArgumentError(f"Failed to load specfile: {err}") from None

        try:
            self.repository = pygit2.Repository(str(self.git_path))
        except pygit2.GitError:
            raise InvalidArgumentError(
                f"{self.git_path} is not in a git repository"
            ) from None
        self._guess_last_ref(baseversion, pre)

        self.evr_only: bool = evr_only

        self.last_entry: GitLogEntry | None = None

    @classmethod
    def make_parser(
        cls, parser_func: Callable = argparse.ArgumentParser, standalone=False, **kwargs
    ) -> argparse.ArgumentParser:
        del standalone
        parser = parser_func(**kwargs)
        parser.add_argument("specpath", nargs="?", type=Path)
        parser.add_argument("-r", "--last-ref", dest="baseversion", metavar="LAST_REF")
        parser.add_argument("--pre")
        parser.add_argument("-o", "--outdir", default=Path.cwd(), type=Path)
        parser.add_argument("--stdout", action="store_true")
        parser.add_argument("-A", "--archive", action="store_true")
        parser.add_argument(
            "--evr-only",
            action="store_true",
            help="Print out the generated evr and exit",
        )
        return parser

    def _guess_last_ref(self, baseversion: str | None, pre: str | None) -> None:
        self.pre = bool(pre)
        if not baseversion:
            try:
                baseversion = self.repository.describe(
                    self.repository.head, abbreviated_size=0
                )
            except pygit2.GitError:
                self.pre = True
                self.baseversion = pre or "0.1.0"
                self.last_ref_hash = next(
                    self.repository.walk(
                        self.repository.head.target, pygit2.GIT_SORT_REVERSE
                    )
                ).id
                return

        if TYPE_CHECKING:
            assert isinstance(baseversion, str)  # satisfy mypy
        try:
            ref = self.repository.revparse_single(baseversion)
        except KeyError:
            raise InvalidArgumentError(f"Invalid baseversion {baseversion!r}") from None

        self.last_ref_hash = ref.target if hasattr(ref, "target") else ref.id
        if pre:
            self.baseversion = pre
        elif isinstance(ref, pygit2.Tag):
            self.baseversion = baseversion.lstrip("v")
        else:
            self.baseversion = baseversion

    def write_spec(self) -> None:
        content = str(self.spec)
        if self.stdout:
            sys.stdout.write(content)
            return None
        out = self.outdir.joinpath(self.specpath.name)
        try:
            out.write_text(content)
        except OSError as err:
            sys.exit(f"Failed to output specfile to {out}: {err}")

    def add_entries(self) -> None:
        walker = self.repository.walk(
            self.repository.head.target, pygit2.GIT_SORT_REVERSE
        )
        walker.hide(self.last_ref_hash)
        with self.spec.changelog() as changelog:
            for index, commit in enumerate(walker):
                self.last_entry = GitLogEntry(
                    commit=commit,
                    index=index + 1,
                    baseversion=self.baseversion,
                    pre=self.pre,
                )
                self.last_entry.add_to_clog(changelog)

    def create_archive(self, nv: str):
        with tarfile.open(self.outdir / f"{nv}.tar.gz", "w:gz") as tf:
            self.repository.write_archive(
                self.repository.head.target, tf, prefix=f"{nv}/"
            )

    def _reference_archive(self, nv: str) -> None:
        archive_name = nv + ".tar.gz"
        with self.spec.sources() as sources:
            sources[0].location = archive_name
        with self.spec.prep() as prep:
            for name in ("autosetup", "setup"):
                if "%" + name not in prep:
                    continue
                macro = getattr(prep, name)
                del macro.options.n
                # Clear forgesetupargs
                tokens = getattr(macro.options, "_tokens", None)
                if tokens is None:
                    break
                for idx, token in enumerate(tokens):
                    if (
                        "forgesetupargs" in token.value
                        and token.type == options.TokenType.DEFAULT
                    ):
                        del tokens[idx]
                        break
                break

    def run(self) -> int:
        self.add_entries()
        version = self.last_entry.version if self.last_entry else self.spec.version
        if self.evr_only:
            print(version)
            return 0
        if self.last_entry:
            self.spec.set_version_and_release(version, "1")
        nv = f"{self.spec.expanded_name}-{self.spec.expanded_version}"
        if self.archive:
            self.create_archive(nv)
            self._reference_archive(nv)
        self.write_spec()
        return 0


class DevSRPM(DevEntries):
    def __init__(
        self,
        *,
        specpath: Path | None,
        baseversion: str | None,
        pre: str | None,
        # Unique
        srpm_outdir: Path,
        keep: bool,
        clean_srpms: bool,
        # Unused
        evr_only: bool = False,  # noqa: ARG002
    ) -> None:
        self.keep: bool = keep
        self.srpm_outdir = srpm_outdir
        self.clean_srpms = clean_srpms
        if self.keep:
            outdir = self.srpm_outdir
        else:
            outdir = Path(tempfile.mkdtemp())
            self.cleanup.append(
                functools.partial(shutil.rmtree, outdir, ignore_errors=True)
            )

        super().__init__(
            specpath=specpath,
            baseversion=baseversion,
            outdir=outdir,
            pre=pre,
            archive=True,
            stdout=False,
        )

    def run(self) -> int:
        if r := super().run():
            return r
        try:
            self.build_srpm()
        except subprocess.CalledProcessError as err:
            LOG.error("Failed to run: %s", err.cmd)
            return err.returncode
        if self.clean_srpms:
            for file in self.srpm_outdir.glob(f"{self.spec.expanded_name}-*.src.rpm"):
                if self.spec.expanded_version not in file.name:
                    LOG.debug("Removing old SRPM %r", str(file))
                    file.unlink()
        return 0

    def build_srpm(self) -> subprocess.CompletedProcess:
        defines = {
            # "_topdir": self.outdir,
            "_sourcedir": self.outdir,
            "_specdir": self.outdir,
            "_srcrpmdir ": self.srpm_outdir,
        }
        cmd: list[StrOrBytesPath] = [
            "rpmbuild",
            "-bs",
            self.outdir / self.specpath.name,
        ]
        for name, value in defines.items():
            cmd.extend(("-D", f"{name} {value}"))
        LOG.info("Building SRPM: %s", cmd)
        proc = subprocess.run(cmd, check=True)
        return proc

    @classmethod
    def make_parser(
        cls, parser_func: Callable = argparse.ArgumentParser, standalone=False, **kwargs
    ) -> argparse.ArgumentParser:
        del standalone
        parser = parser_func(**kwargs)
        parser.add_argument("specpath", nargs="?", type=Path)
        parser.add_argument("-r", "--last-ref", dest="baseversion", metavar="LAST_REF")
        parser.add_argument("--pre")
        parser.add_argument(
            "-o",
            "--outdir",
            default=Path.cwd(),
            type=Path,
            dest="srpm_outdir",
            metavar="OUTDIR",
        )
        parser.add_argument(
            "--clean",
            action="store_true",
            help="Cleanup old SRPMs in --outdir",
            dest="clean_srpms",
        )
        parser.add_argument("-k", "--keep", action="store_true")
        return parser



# Copyright (c) 2023 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import datetime
import subprocess
from collections.abc import Sequence
from pathlib import Path
from shutil import copy2
from unittest import mock

import pytest
import pytest_mock
import specfile
import specfile.changelog
import specfile.macros

from fclogr.cli import Bumper, main

DEFAULT_PACKAGER = "Perry the Packager <perry@example.com>"
DATE_STR = datetime.datetime.now(tz=datetime.timezone.utc).strftime("%a %b %d %Y")


@pytest.mark.parametrize(
    "value,expected",
    [
        # REL_MATCHER
        pytest.param("1", "2", id="simple"),
        pytest.param("33", "34", id="simple double"),
        # PRE_REL_MATCHER
        pytest.param("0.1", "0.2", id="prerel"),
        # POST_REL_MATCHER
        pytest.param("abcdefgh.1", "abcdefgh.2", id="postrel"),
        # fallback
        pytest.param("abcdefgh", "abcdefgh.1", id="fallback"),
    ],
)
def test_bumper_handle_release(value: str, expected: str) -> None:
    assert Bumper._handle_release2(value) == expected


def get_entry_obj(
    evr: str,
    lines: list[str],
    following_lines: list[str],
    date_str: str = DATE_STR,
    packager: str = DEFAULT_PACKAGER,
):
    header = f"* {date_str} {packager} - {evr}"
    return specfile.changelog.ChangelogEntry(header, lines, following_lines)


@pytest.fixture()
def rpm_packager(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("RPM_PACKAGER", DEFAULT_PACKAGER)


INITIAL_ENTRIES = (
    get_entry_obj(
        "1-1",
        ["- Initial package"],
        [],
        "Fri Mar 03 2023",
        "Packager <example@example.com>",
    ),
    get_entry_obj(
        "1-2",
        ["- rebuilt"],
        [""],
        "Fri Mar 03 2023",
        "Packager <example@example.com>",
    ),
)


@pytest.mark.parametrize(
    "args, changelog, version, raw_release, subprocess_calls",
    [
        pytest.param(
            [],
            [*INITIAL_ENTRIES, get_entry_obj("1-3", ["- bump"], [""])],
            "1",
            "3%{?dist}",
            [],
            id="no-arguments",
        ),
        pytest.param(
            ["--new", "3"],
            [*INITIAL_ENTRIES, get_entry_obj("3-1", ["- Update to 3."], [""])],
            "3",
            "1%{?dist}",
            [],
            id="new-version-no-arguments",
        ),
        pytest.param(
            ["--new", "3", "--commit"],
            [*INITIAL_ENTRIES, get_entry_obj("3-1", ["- Update to 3."], [""])],
            "3",
            "1%{?dist}",
            [
                lambda p: mock.call(
                    ["add", p],
                    check=True,
                    stdout=subprocess.DEVNULL,
                ),
                lambda _: mock.call(["commit", "-m", "Update to 3."], check=True),
            ],
            id="new-version-commit",
        ),
        pytest.param(
            ["--comment", "abc"],
            [*INITIAL_ENTRIES, get_entry_obj("1-3", ["- abc"], [""])],
            "1",
            "3%{?dist}",
            [],
            id="comment",
        ),
        pytest.param(
            ["--comment", "%changelog is escaped"],
            [
                *INITIAL_ENTRIES,
                get_entry_obj("1-3", ["- %%changelog is escaped"], [""]),
            ],
            "1",
            "3%{?dist}",
            [],
            id="comment-escape",
        ),
        pytest.param(
            ["--comment", "- abc", "-c", "xyz", "--new", "5", "-S"],
            [*INITIAL_ENTRIES, get_entry_obj("5-1", ["- abc", "- xyz"], [""])],
            "5",
            "1%{?dist}",
            [
                lambda p: mock.call(
                    ["add", p],
                    check=True,
                    stdout=subprocess.DEVNULL,
                ),
                lambda _: mock.call(
                    ["commit", "-m", "abc", "-m", "xyz", "--gpg-sign"], check=True
                ),
            ],
            id="new-version-with-multiple-comments",
        ),
    ],
)
def test_bumper_full(
    test_data: Path,
    tmp_path: Path,
    rpm_packager: None,
    mocker: pytest_mock.MockerFixture,
    #
    args: Sequence[str],
    changelog: list[specfile.changelog.ChangelogEntry],
    version: str,
    raw_release: str,
    subprocess_calls: Sequence[mock._Call],
):
    subprocess_mock: mock.MagicMock = mocker.patch.object(Bumper, "_git")
    name = "package.2.spec"
    path = tmp_path / name
    copy2(test_data / name, path)
    args = ["bump", *args, str(path)]
    assert main(args) == 0
    specfile.macros.Macros.reinit()

    with specfile.Specfile(path) as spec:
        assert spec.raw_release == raw_release
        assert spec.version == version
        with spec.changelog() as spec_changelog:
            assert list(spec_changelog) == changelog
    assert subprocess_mock.call_args_list == [call(path) for call in subprocess_calls]


@pytest.mark.parametrize(
    "args,subprocess_calls,version",
    [
        pytest.param([], [], None, id="simple-noop"),
        pytest.param(["--new", "10"], [], "10", id="new"),
        pytest.param(
            ["--new", "10", "--commit"],
            [
                lambda p: mock.call(
                    ["add", p],
                    check=True,
                    stdout=subprocess.DEVNULL,
                ),
                lambda _: mock.call(
                    ["commit", "-m", "Update to 10.", "--allow-empty"], check=True
                ),
            ],
            "10",
            id="new-commit",
        ),
    ],
)
def test_bumper_autospec(
    test_data: Path,
    tmp_path: Path,
    mocker: pytest_mock.MockerFixture,
    #
    args: Sequence[str],
    version: str | None,
    subprocess_calls,
) -> None:
    subprocess_mock: mock.MagicMock = mocker.patch.object(Bumper, "_git")
    name = "package-autospec.spec"
    src = test_data / name
    dest = tmp_path / name
    copy2(src, dest)

    args = ["bump", *args, str(dest)]
    assert main(args) == 0
    if version is None:
        assert src.read_text() == dest.read_text()
    else:
        specfile.macros.Macros.reinit()
        with specfile.Specfile(dest) as spec:
            assert spec.expanded_version == version
            spec.update_tag("Version", "1")
            assert spec.expanded_version == "1"
            assert src.read_text() == str(spec)
    assert subprocess_mock.call_args_list == [call(dest) for call in subprocess_calls]


def replace_instance(lines: Sequence[str], match: str, repl: str) -> list[str]:
    lines = list(lines)
    for idx, line in enumerate(lines):
        if match in line:
            line = line.replace(match, repl)
            lines[idx] = line
            break
    return lines


@pytest.mark.parametrize(
    "initial, final, command, replace_date",
    [
        pytest.param(
            "package-forge.spec",
            "package-forge.2.spec",
            ["--new", "0.60.0"],
            "Thu Feb 01 2024",
            id="package-forge",
        ),
        pytest.param(
            "package-forgeversion.spec",
            "package-forgeversion.2.spec",
            ["--new", "0.60.0"],
            "Thu Feb 01 2024",
            id="package-forge",
        ),
    ],
)
def test_bumper_forge_compare(
    test_data: Path,
    tmp_path: Path,
    rpm_packager: None,
    initial: str,
    final: str,
    command: Sequence[str],
    replace_date: str,
) -> None:
    if (
        "forgeversion" in initial
        and subprocess.run(
            ["rpm", "-E", "%{defined forgeversion}"],
            check=False,
            capture_output=True,
            text=True,
        ).stdout.strip()
        == "0"
    ):
        pytest.skip("forgeversion is not supported in RHEL!")
    src = test_data / initial
    dest = tmp_path / initial
    copy2(src, dest)
    final_path = test_data / final
    final_lines = replace_instance(
        final_path.read_text().splitlines(), DATE_STR, replace_date
    )

    args: list[str] = ["bump", str(dest), *command]
    assert main(args) == 0

    dest_lines = replace_instance(dest.read_text().splitlines(), DATE_STR, replace_date)
    assert final_lines == dest_lines



# SPDX-FileCopyrightText: 2023 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later


from __future__ import annotations

import argparse
import logging
import re
import subprocess
from collections.abc import Callable, Sequence
from pathlib import Path
from typing import TYPE_CHECKING

import specfile
import specfile.changelog
import specfile.macro_definitions

from fclogr._util import escape_percentage
from fclogr.cli.base import Command, InvalidArgumentError

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath

# https://pagure.io/fedora-infra/rpmautospec/blob/edf18a1967f05253f937aa93ea90209459d18695/f/rpmautospec/misc.py#_11
AUTORELEASE_RE = re.compile(
    r"%(?:autorelease(?:\s|$)|\{\??autorelease(?:\s+[^\}]*)?\})"
)
#
PRE_REL_MATCHER = re.compile(r"^(0\.)(\d+)(.*)")
REL_MATCHER = re.compile(r"^(\d+)(.*)")
POST_REL_MATCHER = re.compile(r"^([^\d]+\.)(\d+)($)")

RELEASE_PATTERNS: tuple[re.Pattern[str], ...] = (
    PRE_REL_MATCHER,
    REL_MATCHER,
    POST_REL_MATCHER,
)

LOG = logging.getLogger(__name__)


class Bumper(Command):
    """
    Bump the release and changelog of a specfile.
    """

    def __init__(
        self,
        *,
        specpath: Path | None,
        comment: list[str] | None,
        version: str | None,
        entry_only: bool,
        commit: bool,
        gpg_sign: bool,
    ) -> None:
        self.specpath: Path = self._v_specpath(specpath)
        self.version: str | None = version
        self.entry_only: bool = entry_only
        self.commit: bool = gpg_sign or commit
        self.gpg_sign: bool = gpg_sign
        try:
            self.spec = specfile.Specfile(self.specpath, force_parse=True)
        except specfile.exceptions.SpecfileException as err:
            LOG.debug("Failed to load specfile", exc_info=True)
            raise InvalidArgumentError(f"Failed to load specfile: {err}") from None
        comment = comment or [f"Update to {self.version}." if self.version else "bump"]
        self.comment: list[str] = [
            escape_percentage(c if c.startswith("- ") else "- " + c) for c in comment
        ]
        if entry_only and version:
            raise InvalidArgumentError("--entry-only and --new are mutually exclusive.")

    def _write_spec(self) -> int | str:
        try:
            self.specpath.write_text(str(self.spec), encoding="utf-8")
        except OSError as err:
            return f"Failed to output specfile to {self.specpath}: {err}"
        return 0

    def _evr(self) -> str:
        return self.spec.expand("%{?epoch:%{epoch}:}%{version}-%{release}")

    def _bump_version(self, version: str) -> None:
        macros: specfile.macro_definitions.MacroDefinitions
        with self.spec.macro_definitions() as macros:
            if "version0" in macros:
                self._update_macro("version0", version, macros)
                return
        self.spec.update_tag("Version", version)

    def run(self) -> int | str:
        uses_autorelease: bool = False
        old_evr = self._evr()

        if self.version:
            self._bump_version(self.version)

        if self.entry_only:
            pass
        elif any(AUTORELEASE_RE.search(line) for line in str(self.spec).splitlines()):
            uses_autorelease = True
            LOG.info("%s uses %%autorelease. Preserving Release.", self.specpath)
        elif self.version:
            self.spec.release = "1"
        else:
            self._handle_release()

        if not uses_autorelease:
            LOG.info("%s: bumped %s -> %s", self.specpath, old_evr, self._evr())
        elif self.version:
            LOG.info(
                "%s: bumped Version to %s", self.specpath, self.spec.expanded_version
            )

        if not self.spec.has_autochangelog:
            if uses_autorelease:
                LOG.warning(
                    "%s uses %%autorelease, but it does not use %%autochangelog."
                    " We're adding a changelog entry,"
                    " but its EVR part may be incorrect.",
                    self.specpath,
                )
            self.spec.add_changelog_entry(self.comment)
        else:
            LOG.info("%s uses %%autochangelog. Preserving %%changelog.", self.specpath)
        if r := self._write_spec():
            return r
        if self.commit:
            self.git_commit(
                allow_empty=uses_autorelease and self.spec.has_autochangelog
            )
        return 0

    def _update_macro(
        self,
        name: str,
        value: str,
        macros: specfile.macro_definitions.MacroDefinitions,
    ) -> None:
        macro = macros.get(name)
        macro.body = value

    def _handle_release(self) -> None:
        baserelease = self.spec.expand("%{baserelease}")
        if baserelease != "%{baserelease}":
            with self.spec.macro_definitions() as macros:
                self._update_macro(
                    "baserelease", self._handle_release2(baserelease), macros
                )
        else:
            self.spec.release = self._handle_release2(self.spec.release)

    @staticmethod
    def _handle_release2(value: str) -> str:
        for pattern in RELEASE_PATTERNS:
            if match := pattern.match(value):
                groups = list(match.groups())
                groups[-2] = str(int(groups[-2]) + 1)
                return "".join(groups)
        return value + ".1"

    def _get_git_messages(self) -> list[str]:
        args: list[str] = []
        for entry in self.comment:
            entry = re.sub("^- ", "", entry)
            args.extend(("-m", entry))
        return args

    def _git(
        self, cmd: Sequence[StrOrBytesPath], *args, **kwargs
    ) -> subprocess.CompletedProcess:
        cmd = ["git", *cmd]
        LOG.info("Running %s", cmd)
        return subprocess.run(cmd, *args, **kwargs)  # noqa: PLW1510

    def git_commit(self, allow_empty: bool = False) -> None:
        args: list[str | Path] = ["add", self.specpath]
        self._git(args, check=True, stdout=subprocess.DEVNULL)
        args = ["commit", *self._get_git_messages()]
        if allow_empty:
            args.append("--allow-empty")
        if self.gpg_sign:
            args.append("--gpg-sign")
        self._git(args, check=True)

    @classmethod
    def make_parser(
        cls, parser_func: Callable = argparse.ArgumentParser, standalone=False, **kwargs
    ) -> argparse.ArgumentParser:
        kwargs["description"] = cls.__doc__
        if not standalone:
            kwargs["help"] = cls.__doc__
        parser = parser_func(**kwargs)
        parser.add_argument("-c", "--comment", action="append")
        parser.add_argument(
            "--new", dest="version", help="Bump Release to 1 and change Version"
        )
        parser.add_argument(
            "-M",
            "--entry-only",
            action="store_true",
            # help="Don't bump Release",
            help=argparse.SUPPRESS,
        )
        parser.add_argument(
            "--commit", action="store_true", help="Commit the specfile bump"
        )
        parser.add_argument(
            "-S",
            "--gpg-sign",
            action="store_true",
            help="Sign the specfile bump commit." " Implies --commit.",
        )
        parser.add_argument("specpath", nargs="?", type=Path)
        return parser
